/*
 * coleco.c
 *
 * ColecoVision emulation.
 */

/* $Id: coleco.c,v 1.3 1999/07/01 01:26:34 nyef Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "ui.h"
#include "system.h"
#include "cal.h"
#include "tool.h"
#include "tms9918.h"
#include "sms_psg.h"

#define CV_RAMSIZE 0x2400 /* 0x2000 BIOS rom + 0x400 ram */

unsigned char *cv_memory;
unsigned char *cv_rom;
cal_cpu cv_cpu;
tms9918 cv_vdp;

unsigned char *cv_rombase;

const struct joypad_template cv_joypad_template = {
    6, {0x01, 0x02, 0x04, 0x08, 0x10, 0x20}
};

struct joypad cv_joypad_1 = {
    &cv_joypad_template, 0
};

int cv_joypad_state;

unsigned char cv_read(cal_cpu cpu, unsigned long addr)
{
    if (addr < 0x2000) {
	return cv_rom[addr];
    } else if (addr < 0x6000) {
	deb_printf("cv_read: read unassigned address 0x%04x.\n", addr);
	return 0;
    } else if (addr < 0x8000) {
	return cv_memory[addr & 0x3ff];
    } else {
	return romimage[addr & 0x7fff];
    }
}

void cv_write(cal_cpu cpu, unsigned long addr, unsigned char data)
{
    if ((addr & 0xe000) == 0x6000) {
	cv_memory[addr & 0x3ff] = data;
    }
}

memread8_t cv_readfunc = &cv_read;
memwrite8_t cv_writefunc = &cv_write;

unsigned char cv_io_read(cal_cpu cpu, unsigned long address)
{
    if ((address & 0xe0) == 0xa0) {
	if (address & 1) {
	    return cv_vdp->readport1(cv_vdp);
	} else {
	    return cv_vdp->readport0(cv_vdp);
	}
    } else if ((address & 0xe0) == 0xe0) {
	if (address & 0x02) {
	    /* FIXME: controller 2? */
	    return 0xff;
	} else {
	    ui_update_joypad(&cv_joypad_1);
	    if (cv_joypad_state) {
		/* joystick mode */
		return ((cv_joypad_1.data & 0x0f) | ((cv_joypad_1.data & 0x10)? 0x40: 0)) ^ 0xff;
	    } else {
		/* keypad mode */
		/* FIXME: add keypad */
		return ((cv_joypad_1.data & 0x20)? 0x40: 0) ^ 0xff;
	    }
	}
    } else {
	deb_printf("cv_io_read: 0x%02x.\n", address & 0xff);
    }

    return 0;
}

void cv_io_write(cal_cpu cpu, unsigned long address, unsigned char data)
{
    if ((address & 0xe0) == 0x80) {
	/* set controls to keypad mode */
	cv_joypad_state = 0;
    } else if ((address & 0xe0) == 0xa0) {
	if (address & 1) {
	    cv_vdp->writeport1(cv_vdp, data);
	} else {
	    cv_vdp->writeport0(cv_vdp, data);
	}
    } else if ((address & 0xe0) == 0xc0) {
	/* set controls to joystick mode */
	cv_joypad_state = 1;
    } else if ((address & 0xe0) == 0xe0) {
	sms_psg_write(data);
    } else {
	deb_printf("cv_io_write: 0x%02x = 0x%02x.\n", address & 0xff, data);
    }
}

void cv_timeslice(cal_cpu cpu)
{
    int i;

    for (i = 262; i && (system_flags == F_NONE); i--) {
	cpu->runfor(cpu, 228); /* FIXME: this is .5 over what it should be. */
	cpu->run(cpu);

	if (cv_vdp->periodic(cv_vdp)) {
	    cpu->nmi(cpu);
	}
	if (system_flags & F_UNIMPL) {
	    deb_printf("Unimplimented Instruction.\n");
	    system_flags |= F_QUIT;
	}
    }
}

void cv_shutdown(void)
{
    sms_psg_done();
}

void cv_run(int startup_deb)
{
    cv_memory = calloc(1, CV_RAMSIZE);

    if (cv_memory && romimage) {
	cv_rom = cv_memory + 0x400;

	{
	    FILE *romfile;
	    
	    deb_printf("attempting to load bios file.\n");
	    
	    romfile = fopen("coleco.rom", "rb");
	    
	    if (romfile != NULL) {
		fread(cv_rom, 0x4000, 1, romfile);
		fclose(romfile);
	    } else {
		deb_printf("could not open bios file.\n");
		deb_printf("things are about to get ugly.\n");
	    }
	}
	
	cv_cpu = cal_create(CPUT_MARATZ80);
	cv_cpu->setmmu8(cv_cpu, 0, 0, &cv_readfunc, &cv_writefunc);
	cv_cpu->setiou(cv_cpu, cv_io_read, cv_io_write);
	cv_cpu->reset(cv_cpu);

	cv_vdp = tms9918_create();

	dn_shutdown = cv_shutdown;
	
	nes_psg_quality = 2;
	sms_psg_init();

	if (ui_register_joypad(&cv_joypad_1)) {
	    deb_printf("cv_run: joypad 1... registered.\n");
	} else {
	    deb_printf("cv_run: joypad 1... not registered.\n");
	}
	
	set_timeslice((void (*)(void *)) &cv_timeslice, cv_cpu);
    } else {
	deb_printf("out of memory.\n");
    }
}

/*
 * $Log: coleco.c,v $
 * Revision 1.3  1999/07/01 01:26:34  nyef
 * added code to shutdown the sound system
 *
 * Revision 1.2  1999/06/12 00:18:50  nyef
 * added preliminary joypad support
 *
 * Revision 1.1  1999/06/08 01:47:26  nyef
 * Initial revision
 *
 */
