/*
 * debug.c
 *
 * core debugger code.
 */

/* $Id: debug.c,v 1.8 1999/05/02 18:18:14 nyef Exp $ */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>

#include "video.h"
#include "dt6502.h"
#include "emu6502.h"
#include "flags.h"
#include "tool.h"
#include "ui.h"
#include "nes_ppu.h"
#include "debug.h"

int deb_address; /* current IP to disassemble from */
int dump_address; /* current memory location to dump from */

void (*display_header)(void);
void (*display_vectors)(void);
unsigned char (*mem_read_byte)(unsigned short address);
void (*trace)(void *);
void (*freerun)(void *);
int (*disassemble)(int start_addr, int end_addr);

void deb_help(void)
{
    deb_printf("commands:\n");
    deb_printf("  Q: Quit     ?: Help     T: Trace instruction\n");
    deb_printf("  U: Unassemble [start-address] [end-address]\n");
    deb_printf("  R: Registers [regname]      V: Print vectors\n");
    deb_printf("  N: Simulate NMI      I: Simulate IRQ\n");
    deb_printf("  H: Print ROM Header    G: Go\n");
    deb_printf("  P: Print tile pattern\n");
    deb_printf("  D: Dump memory [start-address] [end-address]\n");
}

void dump_flags(char *buf, unsigned char flags)
{
    char *bufptr = buf;
  
    *bufptr++ = (flags & 0x80) ? 'N' : 'n';
    *bufptr++ = ' ';
    *bufptr++ = (flags & 0x40) ? 'V' : 'v';
    *bufptr++ = ' ';
    *bufptr++ = (flags & 0x08) ? 'D' : 'd';
    *bufptr++ = ' ';
    *bufptr++ = (flags & 0x04) ? 'I' : 'i';
    *bufptr++ = ' ';
    *bufptr++ = (flags & 0x02) ? 'Z' : 'z';
    *bufptr++ = ' ';
    *bufptr++ = (flags & 0x01) ? 'C' : 'c';
    *bufptr = '\0';
}

void dump_regs(struct emu6502_context *regs)
{
    char buf[12];
    deb_printf("A=%02X  S=%02X  P=%02X  PC=%04hX\n",
	   regs->reg_A, regs->reg_S, flags_to_6502[regs->reg_P], regs->reg_PC);
    dump_flags(buf, flags_to_6502[regs->reg_P]);
    deb_printf("X=%02X  Y=%02X   %s\n", regs->reg_X, regs->reg_Y, buf);
    disassemble(regs->reg_PC, regs->reg_PC);
    deb_address = regs->reg_PC;
}

void deb(void *context)
{
    char inbuf[128];
    char *inptr;
    struct emu6502_context *regs = context;

    video_enter_deb();

    deb_address = regs->reg_PC;
    dump_address = 0; /* this is somewhat arbitrary. */
    
    dump_regs(regs);
    
    while (!(system_flags & F_QUIT)) {
	deb_printf("->");
	fgets(inbuf, 127, stdin);
	if (inbuf[strlen(inbuf)-1] == '\n') {
	    inbuf[strlen(inbuf)-1] = '\0';
	}
	inptr = inbuf+1;
	if (!((*inptr == ' ') || (*inptr == '\t') || (*inptr == '\0'))) {
	    if (*inbuf != '\0') {
		*inbuf = 0xff;
	    }
	}
	switch (toupper(*inbuf)) {
    
	case 'Q': /* Quit */
	    system_flags |= F_QUIT;
	    break;
      
	case 'V': /* Vectors */
	    display_vectors();
	    break;
      
	case 'H': /* Header */
	    display_header();
	    break;
      
	case '?': /* Help */
	    deb_help();
	    break;

	case 'U': { /* Unassemble */
	    int addr;
	    int max_addr;
	    int count;
      
	    while (*inptr == ' ') inptr++;
	    if (*inptr == '\0') {
		addr = deb_address;
	    } else {
		addr = strtol(inptr, &inptr, 16);
	    }
	    while (*inptr == ' ') inptr++;
	    if (*inptr == '\0') {
		max_addr = addr + 20;
	    } else {
		max_addr = strtol(inptr, &inptr, 16);
		if (max_addr < addr) max_addr = addr;
	    }
	    count = disassemble(addr, max_addr);
	    deb_address = addr + count;
	    break;
	}

	case 'D': { /* Dump memory */
	    int addr;
	    int max_addr;
	    int count;
	    
	    while (*inptr == ' ') inptr++;
	    if (*inptr == '\0') {
		addr = dump_address;
	    } else {
		addr = strtol(inptr, &inptr, 16);
	    }
	    while (*inptr == ' ') inptr++;
	    if (*inptr == '\0') {
		max_addr = addr + 0x7f;
	    } else {
		max_addr = strtol(inptr, &inptr, 16);
	    }
	    max_addr &= 0xffff;
	    addr &= 0xffff;
	    
	    if (addr & 7) {
		deb_printf("%04x:", (addr & 0xfff8));
		for (count = 0; count < (addr & 7); count++) {
		    deb_printf("   ");
		}
		if ((addr & 0xfff8) == (max_addr & 0xfff8)) {
		    for (count = addr; count <= max_addr; count++) {
			deb_printf(" %02x", mem_read_byte(count));
		    }
		} else {
		    for (count = addr; count <= (addr | 0x7); count++) {
			deb_printf(" %02x", mem_read_byte(count));
		    }
		}
		deb_printf("\n");
		addr = (addr + 7) & 0xfff8;
	    }
	    while ((addr & 0xfff8) != (max_addr & 0xfff8)) {
		deb_printf("%04x: %02x %02x %02x %02x %02x %02x %02x %02x\n", addr,
		       mem_read_byte(addr), mem_read_byte(addr+1),
		       mem_read_byte(addr+2), mem_read_byte(addr+3),
		       mem_read_byte(addr+4), mem_read_byte(addr+5),
		       mem_read_byte(addr+6), mem_read_byte(addr+7));
		addr = (addr + 8) & 0xffff;
	    }
	    if (addr != (max_addr+1)) {
		deb_printf("%04x:", addr);
		for (count = addr; count <= max_addr; count++) {
		    deb_printf(" %02x", mem_read_byte(count));
		}
		deb_printf("\n");
	    }
	    dump_address = max_addr + 1;
	    break;
	}

#if 0
	case 'P': /* PPU patterns */
	    while (*inptr == ' ') inptr++;
	    if (*inptr == '\0') {
		deb_printf("Pattern #:");
		fgets(inbuf, 127, stdin);
		inptr = inbuf;
	    }
	    PPU_dump_pattern(strtol(inptr, &inptr, 16) << 4);
	    break;
#endif
	    
	case 'R': /* Registers */
	    while (*inptr == ' ') inptr++;
	    if (*inptr == '\0') {
		dump_regs(regs);
	    } else {
		switch (toupper(*inptr)) {
		case 'A':
		    deb_printf("A:");
		    fgets(inbuf, 127, stdin);
		    if (inbuf[strlen(inbuf)-1] == '\n') {
			inbuf[strlen(inbuf)-1] = '\0';
		    }
		    regs->reg_A = strtol(inbuf, &inptr, 16);
		    break;
		case 'X':
		    deb_printf("X:");
		    fgets(inbuf, 127, stdin);
		    if (inbuf[strlen(inbuf)-1] == '\n') {
			inbuf[strlen(inbuf)-1] = '\0';
		    }
		    regs->reg_X = strtol(inbuf, &inptr, 16);
		    break;
		case 'Y':
		    deb_printf("Y:");
		    fgets(inbuf, 127, stdin);
		    if (inbuf[strlen(inbuf)-1] == '\n') {
			inbuf[strlen(inbuf)-1] = '\0';
		    }
		    regs->reg_Y = strtol(inbuf, &inptr, 16);
		    break;
		case 'S':
		    deb_printf("S:");
		    fgets(inbuf, 127, stdin);
		    if (inbuf[strlen(inbuf)-1] == '\n') {
			inbuf[strlen(inbuf)-1] = '\0';
		    }
		    regs->reg_S = strtol(inbuf, &inptr, 16);
		    break;
		case 'P':
		    inptr++;
		    if (toupper(*inptr) == 'C') {
			deb_printf("PC:");
			fgets(inbuf, 127, stdin);
			if (inbuf[strlen(inbuf)-1] == '\n') {
			    inbuf[strlen(inbuf)-1] = '\0';
			}
			regs->reg_PC = strtol(inbuf, &inptr, 16);
		    } else if ((*inptr == ' ') || (*inptr == '\0')) {
			deb_printf("P:");
			fgets(inbuf, 127, stdin);
			if (inbuf[strlen(inbuf)-1] == '\n') {
			    inbuf[strlen(inbuf)-1] = '\0';
			}
			regs->reg_P = flags_to_8088[strtol(inbuf, &inptr, 16)];
		    } else {
			deb_printf("Unknown register\n");
		    }
		    break;
		default:
		}
	    }
	    break;
      
	case 'T': /* trace */
	    trace(regs);
	    dump_regs(regs);
	    break;
      
	case 'G': /* go */
	    video_leave_deb();
	    freerun(regs);
	    video_enter_deb();
	    if (!(system_flags & F_QUIT)) {
		dump_regs(regs);
	    }
	    break;
      
	case 'N': /* NMI */
	    emu6502_NMI(regs);
	    dump_regs(regs);
	    break;
	    
	case 'I': /* IRQ */
	    emu6502_IRQ(regs);
	    dump_regs(regs);
	    break;
      
	case '\0': /* no command */
	    break;
	    
	default:
	    printf("Unrecognized command. type '?' for help.\n");
	}
    }
    
    video_leave_deb();
}

/*
 * $Log: debug.c,v $
 * Revision 1.8  1999/05/02 18:18:14  nyef
 * removed ppu patern display command
 *
 * Revision 1.7  1999/03/05 04:05:59  nyef
 * fixed problem with the warning fix patch
 *
 * Revision 1.6  1999/02/15 03:24:05  nyef
 * added patch from Michael Vance to squash warnings
 *
 * Revision 1.5  1998/08/02 18:54:19  nyef
 * changed every printf to deb_printf for debug console.
 *
 * Revision 1.4  1998/07/14 01:46:43  nyef
 * moved disassemble() out to dt6502.c as dt6502_disassemble()
 * made disassemble() a new proc pointer
 *
 * Revision 1.3  1998/07/14 01:29:21  nyef
 * broke trace code out through a proc pointer.
 *
 * Revision 1.2  1998/07/14 01:25:47  nyef
 * removed SuperTrace and Go commands.
 * chanced FreeRun command to Go.
 *
 * Revision 1.1  1998/07/11 22:16:15  nyef
 * Initial revision
 *
 */
