/*
 * nes.c
 *
 * contains NES specific code and data.
 */

/* $Id: nes.c,v 1.37 1999/07/01 01:26:41 nyef Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "tool.h"
#include "ui.h"
#include "nes_ppu.h"
#include "mappers.h"
#include "system.h"
#include "debug.h"
#include "nes.h"
#include "dt6502.h"
#include "nes_psg.h"
#include "cal.h"
#include "video.h"

const struct joypad_template nes_joypad_template = {
    8, {0x10, 0x20, 0x40, 0x80, 0x01, 0x02, 0x04, 0x08}
};

struct joypad nes_joypad_1 = {
    &nes_joypad_template, 0
};

struct joypad nes_joypad_2 = {
    &nes_joypad_template, 0
};

cal_cpu cpu;

#define NES_RAMSIZE 0x2800    /* Size of System and Cart Ram */

#include "nespal.c"

char *battery_filename;

unsigned char *vromimage = NULL;  /* Cart VRom area */
unsigned char *PPU_memory = NULL; /* PPU Ram */

unsigned char *bank_0 = NULL;     /* System Ram */

unsigned char *bank_6 = NULL;     /* Cart Ram */
unsigned char *bank_8 = NULL;     /* Rom Page */
unsigned char *bank_A = NULL;     /* Rom Page */
unsigned char *bank_C = NULL;     /* Rom Page */
unsigned char *bank_E = NULL;     /* Rom Page */

unsigned char bank8_reg;
unsigned char bankA_reg;
unsigned char bankC_reg;
unsigned char bankE_reg;

int joy1_latch;
int joy2_latch;

memread8_t nes_bank4_read_hook;
memread8_t nes_bank6_read_hook;
memwrite8_t nes_bank4_write_hook;
memwrite8_t nes_bank6_write_hook;

unsigned char nes_mem_read_byte(unsigned short address)
{
    if (address < 0x2000) {
	return (read_byte(bank_0, address & 0x7ff));
    } else if (address < 0x6000) {
	return 0xff; /* debugger can't read IO space */
    } else if (address < 0x8000) {
	return (read_byte(bank_6, address & 0x1fff));
    } else if (address < 0xA000) {
	return (read_byte(bank_8, address & 0x1fff));
    } else if (address < 0xC000) {
	return (read_byte(bank_A, address & 0x1fff));
    } else if (address < 0xE000) {
	return (read_byte(bank_C, address & 0x1fff));
    } else {
	return (read_byte(bank_E, address & 0x1fff));
    }
}

void sync_banks(void)
{
    bank_8 = &(romimage[0x10+(bank8_reg*0x2000)]);
    bank_A = &(romimage[0x10+(bankA_reg*0x2000)]);
    bank_C = &(romimage[0x10+(bankC_reg*0x2000)]);
    bank_E = &(romimage[0x10+(bankE_reg*0x2000)]);
}

void nes_display_header(void)
{
    deb_printf("Mapper #%d\n", read_byte(romimage, 6) >> 4);
    deb_printf("ROM pages (16k): %d\n", cart_rom_size);
    deb_printf("VROM pages (8k): %d\n", cart_vrom_size);
    deb_printf("Mirroring: %s\n", read_byte(romimage, 6) & 1?
	   "Vertical": "Horizontal");
    deb_printf("Battery: %s\n", read_byte(romimage, 6) & 2? "Yes": "No");
    deb_printf("Trainer: %s\n", read_byte(romimage, 6) & 4? "Yes": "No");
    deb_printf("4Screen: %s\n", read_byte(romimage, 6) & 8? "Yes": "No");
}

void nes_display_vectors(void)
{
    deb_printf("nmi vector: %04hx\n", read_word(bank_E, 0x1ffa));
    deb_printf("reset vector: %04hx\n", read_word(bank_E, 0x1ffc));
    deb_printf("irq vector: %04hx\n", read_word(bank_E, 0x1ffe));
}

#if 0
void nes_trace(struct emu6502_context *regs)
{
    system_flags = F_NONE;
    PPU_enter_cpu(regs);
    emu6502_step(regs);
    PPU_leave_cpu(regs);
    if (system_flags & F_UNIMPL) {
	printf("Unimplimented Instruction\n");
    }
}

void nes_freerun(struct emu6502_context *regs)
{
    system_flags = F_NONE;
    while (system_flags == F_NONE) {
	PPU_enter_cpu(regs);
	emu6502_run(regs);
	PPU_leave_cpu(regs);
    }
    if (system_flags & F_UNIMPL) {
	deb_printf("Unimplimented Instruction\n");
    }
}
#endif

unsigned char nes_read_bank_0(cal_cpu cpu, unsigned long addr)
{
    return bank_0[addr & 0x7ff];
}

unsigned char nes_read_bank_2(cal_cpu cpu, unsigned long addr)
{
    return ppu_io_read(nes_ppu_true, addr);
}

unsigned char nes_read_bank_4(cal_cpu cpu, unsigned long addr)
{
    unsigned char retval;

    if (addr < 0x4016) {
	return 0x00; /* FIXME: sound register reads?? */
    } else if (addr == 0x4016) {
	retval = 0x40 | (joy1_latch & 1);
	joy1_latch >>= 1;
	return retval;
    } else if (addr == 0x4017) {
	return 0x40;
    } else if (nes_bank4_read_hook) {
	return nes_bank4_read_hook(cpu, addr);
    } else {
	deb_printf("unknown I/O read (0x%04x), returning 0.\n", addr);
	return 0;
    }
}

unsigned char nes_read_bank_6(cal_cpu cpu, unsigned long addr)
{
    if (nes_bank6_read_hook) {
	return nes_bank6_read_hook(cpu, addr);
    } else {
	return bank_6[addr & 0x1fff];
    }
}

unsigned char nes_read_bank_8(cal_cpu cpu, unsigned long addr)
{
    return bank_8[addr & 0x1fff];
}

unsigned char nes_read_bank_a(cal_cpu cpu, unsigned long addr)
{
    return bank_A[addr & 0x1fff];
}

unsigned char nes_read_bank_c(cal_cpu cpu, unsigned long addr)
{
    return bank_C[addr & 0x1fff];
}

unsigned char nes_read_bank_e(cal_cpu cpu, unsigned long addr)
{
    return bank_E[addr & 0x1fff];
}

void nes_write_bank_0(cal_cpu cpu, unsigned long addr, unsigned char data)
{
    bank_0[addr & 0x7ff] = data;
}

void nes_write_bank_2(cal_cpu cpu, unsigned long addr, unsigned char data)
{
    ppu_io_write(nes_ppu_true, addr, data);
}

void nes_write_bank_4(cal_cpu cpu, unsigned long addr, unsigned char data)
{
    if (addr < 0x4010) {
	sound_regs[addr & 15](cpu, addr, data);
    } else if (addr < 0x4014) {
	/* FIXME: more sound regs go here */
    } else if (addr == 0x4014) {
	PPU_do_dma(data);
    } else if (addr == 0x4015) {
	nes_psg_write_control(cpu, addr, data);
    } else if (addr == 0x4016) {
	ui_update_joypad(&nes_joypad_1);
	ui_update_joypad(&nes_joypad_2);
	joy1_latch = nes_joypad_1.data;
	joy2_latch = nes_joypad_2.data;
    } else if (addr == 0x4017) {
	/* FIXME: IRQ control? */
    } else if (nes_bank4_write_hook) {
	nes_bank4_write_hook(cpu, addr, data);
    } else {
	deb_printf("unknown I/O write (0x%04x, 0x%02x).\n", addr, data);
    }
}

void nes_write_bank_6(cal_cpu cpu, unsigned long addr, unsigned char data)
{
    if (nes_bank6_write_hook) {
	nes_bank6_write_hook(cpu, addr, data);
    } else {
	bank_6[addr & 0x1fff] = data;
    }
}

void nes_write_mapper(cal_cpu cpu, unsigned long addr, unsigned char data)
{
    cur_write(addr, data);
}

memread8_t nes_readtable[8] = {
    nes_read_bank_0,
    nes_read_bank_2,
    nes_read_bank_4,
    nes_read_bank_6,
    nes_read_bank_8,
    nes_read_bank_a,
    nes_read_bank_c,
    nes_read_bank_e,
};

memwrite8_t nes_writetable[8] = {
    nes_write_bank_0,
    nes_write_bank_2,
    nes_write_bank_4,
    nes_write_bank_6,
    nes_write_mapper,
    nes_write_mapper,
    nes_write_mapper,
    nes_write_mapper,
};

void nes_timeslice(cal_cpu cpu)
{
    int i;

    for (i = 243; i && (system_flags == F_NONE); i--) {
	if (system_flags == F_NONE) {
/* 	    PPU_enter_cpu(cpu); */
	    cpu->runfor(cpu, 114);
	    nes_psg_enter_cpu(cpu->timeleft(cpu));
	    cpu->run(cpu);
	    if (nesppu_periodic(nes_ppu_true)) {
		cpu->nmi(cpu);
	    }
	    nes_psg_leave_cpu(cpu->timeleft(cpu));
/* 	    PPU_leave_cpu(cpu); */
	    if (system_flags & F_UNIMPL) {
		deb_printf("Unimplimented Instruction\n");
	    }
	}
    }
}

void nes_shutdown(void)
{
    FILE *batfile;
    
    if (read_byte(romimage, 6) & 2) {
	deb_printf("attempting to save battery file.\n");

	/* FIXME: not robust enough. */
	batfile = fopen(battery_filename, "wb");
	
	if (batfile != NULL) {
	    fwrite(bank_6, 0x2000, 1, batfile);
	    fclose(batfile);
	} else {
	    deb_printf("could not open battery file.\n");
	}
    }

    nes_psg_done();
}

void nes_run(int startup_deb)
{
    int mapper;
    
    display_header = &nes_display_header;
    display_vectors = &nes_display_vectors;
    mem_read_byte = &nes_mem_read_byte;
#if 0
    trace = (void (*)(void *))&nes_trace;
    freerun = (void (*)(void *))&nes_freerun;
#endif
    disassemble = &dt6502_disassemble;
    dn_shutdown = &nes_shutdown;
    
    ramblock = calloc(1, NES_RAMSIZE);

    if ((romimage != NULL) && (ramblock != NULL)) {
	if (romimage[0] == 'F') {
	    deb_printf("nes_run(): FDS incomplete, use at own risk.\n");
	    mapper = 20;
	} else {
	    if (cart_is_ext_mapper) {
		deb_printf("using 8-bit mapper id.\n");
	    } else {
		deb_printf("using 4-bit mapper id.\n");
	    }
	    mapper = cart_is_ext_mapper? cart_ext_mapper: cart_mapper;
	}
	if (set_mapper(mapper) == -1) {
	    deb_printf("darcnes: mapper #%d unsupported.\n", mapper);
	} else {
	    bank_0 = ramblock;
	    bank_6 = &ramblock[0x800];

	    if ((romimage[0] == 'N') && (read_byte(romimage, 6) & 2)) {
		FILE *batfile;
		
		deb_printf("attempting to load battery file.\n");

		/* FIXME: this is not the most robust solution I've seen. */
		battery_filename = strdup(romfilename);
		*((unsigned long *) &battery_filename[strlen(battery_filename) - 4]) = 0x7661732e; /* ".sav" */

		batfile = fopen(battery_filename, "rb");

		if (batfile != NULL) {
		    fread(bank_6, 0x2000, 1, batfile);
		    fclose(batfile);
		} else {
		    deb_printf("could not open battery file.\n");
		}

	    }
	    
/* 	    cpu = cal_create(CPUT_EMU6502A); */
	    cpu = cal_create(CPUT_MARAT6502);
	    cpu->setzpage(cpu, ramblock);
	    cpu->setmmu8(cpu, 13, 7, nes_readtable, nes_writetable);
	    PPU_init();
	    if (cart_vrom_size > 0) { /* this works as is for FDS */
		vromimage = &romimage[0x10+(cart_rom_size*0x4000)];
		ppu_flags |= PPUF_VROM;
	    } else {
		vromimage = NULL;
	    }
	    if (ui_register_joypad(&nes_joypad_1)) {
		deb_printf("nes_run: joypad 1... registered.\n");
	    } else {
		deb_printf("nes_run: joypad 1... not registered.\n");
	    }
	    if (ui_register_joypad(&nes_joypad_2)) {
		deb_printf("nes_run: joypad 2... registered.\n");
	    } else {
		deb_printf("nes_run: joypad 2... not registered.\n");
	    }
	    video_setpal(64, nes_palbase_red, nes_palbase_green, nes_palbase_blue);
	    nes_psg_quality = 2;
	    nes_psg_init();
	    cur_init();
	    cpu->reset(cpu);
	    
	    set_timeslice((void (*)(void *)) &nes_timeslice, cpu);
	}
    } else {
	deb_printf("out of memory\n");
    }
}

/*
 * $Log: nes.c,v $
 * Revision 1.37  1999/07/01 01:26:41  nyef
 * added code to shutdown the sound system
 *
 * Revision 1.36  1999/06/19 15:08:44  nyef
 * continued reorganizing the NES PPU
 *
 * Revision 1.35  1999/06/05 02:41:53  nyef
 * converted to use new joypad interface
 * added second joypad support
 *
 * Revision 1.34  1999/05/29 22:31:43  nyef
 * started reorganizing the NES PPU
 *
 * Revision 1.33  1999/05/04 01:41:51  nyef
 * fixed unsupported mapper message to say "darcnes" instead of "tool"
 *
 * Revision 1.32  1999/04/17 20:08:49  nyef
 * fixed for new version of the CAL.
 *
 * Revision 1.31  1999/02/19 02:07:06  nyef
 * added hooks for reading and writing banks 4 and 6
 *
 * Revision 1.30  1999/02/16 01:51:42  nyef
 * removed useless reference to io_nes.h
 *
 * Revision 1.29  1999/02/15 03:25:26  nyef
 * added patch from Michael Vance to squash warnings
 *
 * Revision 1.28  1999/02/14 18:33:09  nyef
 * moved I/O and memory access in from io_nes.S
 *
 * Revision 1.27  1999/02/14 06:39:09  nyef
 * added preliminary support for FDS
 *
 * Revision 1.26  1999/02/14 06:18:37  nyef
 * added lack of support for Famicom Disk System
 *
 * Revision 1.25  1999/02/09 01:50:44  nyef
 * added battery file patch from Michael Vance
 *
 * Revision 1.24  1999/01/05 04:28:57  nyef
 * added a call to set the palette.
 *
 * Revision 1.23  1998/12/31 03:53:27  nyef
 * added support for 8-bit mapper ids.
 *
 * Revision 1.22  1998/12/24 04:55:56  nyef
 * fixed nes_timeslice() to return approximately once per vsync.
 *
 * Revision 1.21  1998/12/21 02:57:47  nyef
 * added preliminary battery save support.
 *
 * Revision 1.20  1998/12/21 00:01:20  nyef
 * changed allocation of ramblock to use calloc(3).
 * This should fix SMB1 under DOS.
 *
 * Revision 1.19  1998/12/18 04:20:19  nyef
 * removed some old code. cenerally cleand up.
 *
 * Revision 1.18  1998/12/11 04:32:38  nyef
 * fixed to be compatable with all 6502 cpu cores available under cal.
 *
 * Revision 1.17  1998/12/06 04:21:23  nyef
 * added code to set psg playback quality.
 *
 * Revision 1.16  1998/11/29 18:28:08  nyef
 * started converting to use the new memory access spec and the CAL.
 *
 * Revision 1.15  1998/10/11 21:46:59  nyef
 * changed to use the emu6502a core. Lolo1 now works.
 *
 * Revision 1.14  1998/08/30 17:06:10  nyef
 * changed the vrom check to use the new ppu_flags system.
 * moved the vrom check to after the call to PPU_init().
 *
 * Revision 1.13  1998/08/26 00:27:32  nyef
 * removed code to set the initial ppu mirroring. (It was already in ppu.c)
 *
 * Revision 1.12  1998/08/25 03:35:34  nyef
 * added code to set the initial ppu mirroring from the header data.
 *
 * Revision 1.11  1998/08/22 16:28:29  nyef
 * added check for setting PPU_use_vrom based on vrom size.
 *
 * Revision 1.10  1998/08/22 00:59:47  nyef
 * added sound hooks to nes_timeslice() and nes_run().
 *
 * Revision 1.9  1998/08/15 01:13:33  nyef
 * added test for unimplimented instruction in nes_timeslice()
 *
 * Revision 1.8  1998/08/02 04:17:26  nyef
 * changed every printf to deb_printf for debug console.
 *
 * Revision 1.7  1998/08/02 04:10:42  nyef
 * added ui.h to the include section and changed nes_timeslice to take
 * a void * instead of a struct emu6502_context *.
 *
 * Revision 1.6  1998/08/01 22:33:22  nyef
 * added nes_timeslice() to facilitate a user interface.
 * changes nes_run() to use nes_timeslice.
 *
 * Revision 1.5  1998/07/26 23:11:19  nyef
 * changed read/write table allocation to allocate extra room for zpage
 *
 * Revision 1.4  1998/07/14 01:44:04  nyef
 * added initializer for disassemble()
 *
 * Revision 1.3  1998/07/14 01:28:22  nyef
 * moved trace code in from debug.c
 *
 * Revision 1.2  1998/07/12 21:25:41  nyef
 * made regs static
 *
 * Revision 1.1  1998/07/11 22:18:09  nyef
 * Initial revision
 *
 */
